#!/usr/bin/env python

from distutils.core import setup

setup(name='GaussOpt',
      version='1.1.3',
      description='Gaussian beam analysis',
      keywords='gaussian optics millimeter terahertz thz',
      packages=['gaussopt',],
      author='John Garrett',
      author_email='garrettj403@gmail.com',
      license='MIT',
      url='https://github.com/garrettj403/GaussOpt/',
      long_description=open('README').read(),
     )
